{{ SumoBot.spin  }}

'' This object contains interfaces to low level drivers that operate the robot
'' The following low interfaces are implemented
''   - 4 Axis Servo Control
''   - Differential PWM Motor Drive
''   - 8 Channel A/D Converter
''   - Sonar Proximity Detector
''   - 8 Channel Digital I/O
''   - 2 Channel Mouse Sensor

'' Also implements several composite virtual interfaces from the low level drivers
''   - Edge Sensors - 2 analog inputs and mouse sensor
''   - IR Proximity - 2 analog inputs
''   - Virtual Sonar Array - Sonar Detector and 1 Servo
''   - Push Sensor - Mouse Sensor

'' Implements state machines for high level control interfaces
''   - Object Guidance and Tracking 
''   - Closed Loop Motion Control


CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 6_000_000
                                                                         
  ANALOG      = 1
  DIGITAL     = 2
  MOUSE       = 3
  DRIVE       = 4
  SONAR       = 5

  MinSpeed      = 48

PUB Init 
  MotorFreq(600)
  SetMotorSpeed(0,0)
    
PUB GetDataPtr(type)
  case type
    0: return @DATA_BEGIN
    1: return @ANALOG_FREQ
    2: return @DIGITAL_CTR
    3: return @MOUSE_CTR
    4: return @MOTOR_CTR
    5: return @SONAR_CTR

  return -1

PUB SetRunMode(mode)
  RUN_MODE := mode
PUB RunMode
  return RUN_MODE
  
{{ Low Level Mouse Interface }}

PUB SetMouseCal(xcal, ycal)
  MOUSE_XCAL := xcal
  MOUSE_YCAL := ycal
  
PUB MouseX
  return MOUSE_X
PUB MouseY
  return MOUSE_Y

PUB MouseAngle
  return MOUSE_X / MOUSE_XCAL
PUB MouseDist
  return MOUSE_Y / MOUSE_YCAL
  
PUB MouseDx
  return MOUSE_DX
PUB MouseDy
  return MOUSE_DY
PUB MouseSquall
  return MOUSE_SQUAL
PUB MouseMin
  return MOUSE_MIN
PUB MouseAvg
  return MOUSE_AVG
PUB MouseMax
  return MOUSE_MAX

PUB SetMouseMode(mode)
  MOUSE_MODE := mode
  
PUB ResetMousePos
  MOUSE_X := 0
  MOUSE_Y := 0
PUB ResetMouseX
  MOUSE_X := 0
PUB ResetMouseY
  MOUSE_Y := 0

{{ Low Level Sonar Interface }}

PUB SonarDist
  return SONAR_DIST
PUB SonarAvg
  return SONAR_AVG
PUB SonarDelta
  return SONAR_DELTA
PUB SonarRate(rate)                            ' Set Averaging Rate
  SONAR_RATE := rate
PUB SonarFreq(freq)                            ' Set Ping Frequency
  freq <#= 20
  SONAR_PERIOD := clkfreq / freq

PUB GetDataBlock(start, end)
  long[start] := @DATA_BEGIN
  long[end] := @DATA_END

PUB GetConfigBlock(start, end)
  long[start] := @CONFIG_BEGIN
  long[end] := @CONFIG_END

{{ Low Level Drive Interface }}


PUB SetMotorSpeed(left, right) 

  MOTOR_MODE |= %0101_0101
  
  if left == 0
    MOTOR_MODE &= %1111_1110
  if right == 0
    MOTOR_MODE &= %1110_1111

  if left < 0
    ||left
    MOTOR_MODE &= %1111_1011
  if right < 0
    ||right
    MOTOR_MODE &= %1011_1111    
      

  MOTOR_FREQ := NextMotorFreq      
  MOTOR_LEFT :=  MotorOffset + (MotorTick * left)   
  MOTOR_RIGHT := MotorOffset + (MotorTick * right)  
  
  MOTOR_CTR := 1

PUB SetMotorCal(lf, rf, lr, rr)
  MOTOR_LFCAL := lf
  MOTOR_RFCAL := rf
  MOTOR_LRCAL := lr
  MOTOR_RRCAL := rr
    
PUB SetDriveSpeed(left, right)
  if left > 0
    left += (MOTOR_LFCAL * left) / 100

  if left < 0
    left -= (MOTOR_LRCAL * left) / 100

  if right > 0
    right += (MOTOR_RFCAL * right) / 100

  if right < 0
    right -= (MOTOR_RRCAL * right) / 100

  if left <> 0 and right == 0
    right := 1

  if right <> 0 and left == 0
    left := 1
    
  SetMotorSpeed(left, right)
  
PUB GetLeftSpeed
  result := (MOTOR_LEFT - MotorOffset) / MotorTick
  if not MOTOR_MODE & %0000_0100
    result *= -1
  
PUB GetRightSpeed
  result := (MOTOR_RIGHT - MotorOffset) / MotorTick
  if not MOTOR_MODE & %0100_0000
    result *= -1

PUB GetMotorFreq
  result := clkfreq / MOTOR_FREQ
    
PUB SetServoPos(servo, pos)
  
  'pos := pos <# 170 #> 10
  'pos := 48000 + pos * 1067
  'pos := Constant(96  * 600)  + Constant(2300 * 96  - 96 * 600) * pos
  
  case servo
    0: PWM0 := pos 
    1: PWM1 := pos 
    2: PWM2 := pos 
    3: PWM3 := pos 
    
  MOTOR_CTR := 1

PUB SetServoAngle(servo, angle) | pos       ' 0 degrees being forward (+90  left and -90 to the right).
  case servo
    0:
      pos :=  angle * PWM0_MCAL + PWM0_BCAL
      PWM0 := (pos <# PWM0_MAX) #> PWM0_MIN
    1:
      pos := angle * PWM1_MCAL + PWM1_BCAL
      PWM1 := pos <# PWM1_MAX #> PWM1_MIN
    2:
      pos := angle * PWM2_MCAL + PWM2_BCAL
      PWM2 := pos <# PWM2_MAX #> PWM2_MIN
    3:
      pos := angle * PWM3_MCAL + PWM3_BCAL
      PWM3 := pos <# PWM3_MAX #> PWM3_MIN
    
  MOTOR_CTR := 1

PUB SetServoCal(servo, mcal, bcal, pmin, pmax)
  case servo
    0:
      PWM0_MCAL := mcal
      PWM0_BCAL := bcal
      PWM0_MIN := pmin
      PWM0_MAX := pmax
    1:
      PWM1_MCAL := mcal
      PWM1_BCAL := bcal
      PWM1_MIN := pmin
      PWM1_MAX := pmax
    2:
      PWM2_MCAL := mcal
      PWM2_BCAL := bcal
      PWM2_MIN := pmin
      PWM2_MAX := pmax
    3:
      PWM3_MCAL := mcal
      PWM3_BCAL := bcal
      PWM3_MIN := pmin
      PWM3_MAX := pmax
      
PUB GetServoPos(servo) | pos
  
  case servo
    0: pos := PWM0    
    1: pos := PWM1
    2: pos := PWM2
    3: pos := PWM3

  result := pos
    

PUB IsServoEnabled(servo)
  case servo
    0: result := SERVO_MODE & %0001
    1: result := SERVO_MODE & %0010
    2: result := SERVO_MODE & %0100
    3: result := SERVO_MODE & %1000

  
PUB EnableServo(servo)
  case servo
    -1: SERVO_MODE |= %1111 
    0: SERVO_MODE |= %0001
    1: SERVO_MODE |= %0010
    2: SERVO_MODE |= %0100
    3: SERVO_MODE |= %1000

  MOTOR_CTR := 1
    
PUB DisableServo(servo)
  case servo
    -1: SERVO_MODE &= %1111_0000 
    0: SERVO_MODE &= %1111_1110
    1: SERVO_MODE &= %1111_1101 
    2: SERVO_MODE &= %1111_1011
    3: SERVO_MODE &= %1111_0111
        
  MOTOR_CTR := 1
  
  
PUB MotorFreq(freq) | usec
  usec := clkfreq / freq
  MotorPeriod(usec)
   
PRI MotorPeriod(usec) | minduty

  minduty := usec * MOTOR_MINDUTY / 100
  
  NextMotorFreq := usec
  MotorTick := (usec - minduty) / MOTOR_SCALE
  MotorOffset := usec - (MOTOR_SCALE * MotorTick)


{{ Low Level I/O Interface }}
PUB SetAnalogEnable(mask)
  ANALOG_EN := mask

PUB SetAnalogMode(mask)
  ANALOG_MODE := mask

PUB SetAnalogRate(rate)
  ANALOG_RATE := rate

PUB SetAnalogFreq(freq)
  ANALOG_FREQ := clkfreq / freq

PUB GetAnalog(channel)
  case channel
    0: result := AN0
    1: result := AN1
    2: result := AN2
    3: result := AN3
    4: result := AN4
    5: result := AN5
    6: result := AN6
    7: result := AN7

PUB GetIR(channel) | raw
  case channel
    5:
      result := AN5_MCAL / AN5 - AN5_BCAL
    6:
      result := AN6_MCAL / AN6 - AN6_BCAL
    7:
      result := AN7_MCAL / AN7 - AN7_BCAL

PUB SetIRCal(channel, mcal, bcal)
  case channel
    5:
      AN5_MCAL := mcal
      AN5_BCAL := bcal
    6:
      AN6_MCAL := mcal
      AN6_BCAL := bcal
    7:
      AN7_MCAL := mcal
      AN7_BCAL := bcal
                  
PUB SetDigitalFreq(freq)
  DIGITAL_FREQ := clkfreq / freq
  
PUB GetDigitalInput
  result := DIG_IN

PUB SetDigitalOutput(val)
  DIG_OUT := val
  
PUB SetDigitalDir(dir)
  DIG_DIR := dir
  
{{ Configuration Data }}
PUB GetCommConfig(port, dest) : addr | sz
  sz := @COM1_RX_PIN - @COM0_RX_PIN
  addr := @COM0_RX_PIN + port * sz
  if dest <> -1
    bytemove(dest, addr, sz)

PUB GetConfigData
  return @COM0_RX_PIN         


PUB GetSonarDistPtr
  return @SONAR_DIST
  
DAT
  DATA_BEGIN   byte 1                           ' Any value other than 0 will cause the EEPROM data to be overwritten

  ANALOG_FREQ  long 96*1000                     ' Analog Sample Frequency
  DIGITAL_FREQ long 96*1000*25                  ' Digital Sample Frequency
                                          
  ANALOG_CTR   long 0                           ' Analog Data Update Counter
  ANALOG_MODE  byte %1110_0011                  ' Channel Mode Average/Raw
  ANALOG_EN    byte %1110_0011                  ' Channel Enable Bits
  ANALOG_RATE  word 25                          ' Averaging Rate

  AN0          word 0                           ' Analog Data 12 bits / Channel
  AN1          word 0
  AN2          word 0
  AN3          word 0
  AN4          word 0
  AN5          word 0
  AN6          word 0
  AN7          word 0

  AN0_SUM      long 0                           ' Temporary Buffer for Averaging
  AN1_SUM      long 0
  AN2_SUM      long 0
  AN3_SUM      long 0
  AN4_SUM      long 0
  AN5_SUM      long 0
  AN6_SUM      long 0
  AN7_SUM      long 0
  
  DIGITAL_CTR  long 0                            ' Digital Update Counter
  DIGITAL_MODE long 0

  DIG_MASK     byte %0000_0000                  ' Channel interrupt on change mask
  DIG_DIR      byte %0000_0000                  ' Channel Direction Input/Output
  DIG_IN       byte %0000_0000                  ' Input Register  
  DIG_OUT      byte %0000_0000                  ' Output Register

  MOUSE_CTR    long 0                           ' Mouse Data Update
  MOUSE_RATE   long 96 * 1000 * 10              ' Rate Sample Period (10 msec)
  
  MOUSE_X      long 0
  MOUSE_Y      long 0
  MOUSE_DX     long 0
  MOUSE_DY     long 0

  MOUSE_SQUAL  byte 0
  MOUSE_MIN    byte 0
  MOUSE_MAX    byte 0
  MOUSE_AVG    byte 0   

  MOUSE_MODE   long %00000_000                  ' Mouse Sensor Control Bits
                                                ' Bit 0 - Enable Sensor
                                                ' Bit 1 - LED Control Mode
  
  MOTOR_CTR   long 0                            ' Motor Update Counter

  MOTOR_MODE   long %0000_0000                   ' Left/Right Motor Mode Control Bits 
                                                '       Bit 0,4 - Enable Motor
                                                '       Bit 1,5 - Dynamic Brake on Stop
                                                '       Bit 2,6 - Forward/Reverse
                                                      

  SERVO_MODE   long %0000_0000                   ' Servo 0-3 PWM Mode Control Bits
                                                '       Bit 0-3 - Enable Servo

  MOTOR_FREQ   long 96*800-9                    'Motor PWM Frequency

  MOTOR_LEFT   long 0
  MOTOR_RIGHT  long 0

  PWM0         long 0
  PWM1         long 0
  PWM2         long 0
  PWM3         long 0

  SONAR_CTR    long 0                            ' Sonar Data Update Counter

  SONAR_PERIOD long 96*50_000-9                  ' Sonar Ping Rate 0 - 20 Hz
  SONAR_DELTA  long 0                            ' Sonar Rate of Change 

  SONAR_DIST   word 0                            ' Current Sonar Ping Distance
  SONAR_RATE   word 5                            ' Averging Rate
  SONAR_AVG    word 0                            ' Average Sonar Distance
  
  DATA_END     byte 0

  CONFIG_BEGIN byte 0                            ' Any value other than 0 will cause the EEPROM data to be overwritten 

  'Serial Port Configuration Parameters
  COMM_PORTS    byte 3                           ' Maximum Number of Communications Ports 
      
  COM0_RX_PIN   long 0                           ' Primary Communications Port 
  COM0_TX_PIN   long 1
  COM0_BAUD     long 19200
  COM0_MODE     long 0
  COM0_OPTIONS  long 0
    
  COM1_RX_PIN   long 31                          ' Debug Communications Port 
  COM1_TX_PIN   long 30
  COM1_BAUD     long 115200
  COM1_MODE     long 0  
  COM1_OPTIONS  long 0
    
  COM2_RX_PIN   long 2                           ' Alternate Communications Port 
  COM2_TX_PIN   long 3
  COM2_BAUD     long 9600
  COM2_MODE     long 0  
  COM2_OPTIONS  long 0

  MOTOR_SCALE   long 1000                        ' Full Scale Motor Value
  MOTOR_MINDUTY long 20                          ' % Minimum Duty Cycle

  MOTOR_LEFT_EN_PIN  long 0
  MOTOR_LEFT_PWM_PIN long 0
  MOTOR_LEFT_DIR0_PIN long 0
  MOTOR_LEFT_DIR1_PIN long 0
          
  MOTOR_RIGHT_EN_PIN   long 0
  MOTOR_RIGHT_PWM_PIN  long 0
  MOTOR_RIGHT_DIR0_PIN long 0
  MOTOR_RIGHT_DIR1_PIN long 0

  AN5_MCAL      long 83000
  AN6_MCAL      long 83000
  AN7_MCAL      long 83000
  AN5_BCAL      long 10
  AN6_BCAL      long 10
  AN7_BCAL      long 10

  PWM0_MCAL     long 1067
  PWM0_BCAL     long 48000
  PWM0_MAX      long 2300 * 96
  PWM0_MIN      long 600 * 96

  PWM1_MCAL     long 1067
  PWM1_BCAL     long 48000
  PWM1_MAX      long 2300 * 96
  PWM1_MIN      long 600 * 96

  PWM2_MCAL     long 1067
  PWM2_BCAL     long 48000
  PWM2_MAX      long 2300 * 96
  PWM2_MIN      long 600 * 96

  PWM3_MCAL     long 1067
  PWM3_BCAL     long 48000
  PWM3_MAX      long 2300 * 96
  PWM3_MIN      long 600 * 96

  MOTOR_LFCAL    long 0
  MOTOR_RFCAL    long 0
  MOTOR_LRCAL    long 0
  MOTOR_RRCAL    long 0

  MOUSE_YCAL     long 1
  MOUSE_XCAL     long 1
  RUN_MODE       long 0
  
  CONFIG_END    word 0

  ' Shared Variables
  MotorTick     long 0
  NextMotorFreq long 0
  MotorOffset   long 0